#ifndef _DS50DB_RESULT
#define _DS50DB_RESULT
#include <string>
#include <map>
#include <boost/lexical_cast.hpp>
#include "cetlib/exception.h"

namespace pqxx {
  class result;
}

namespace ds50 {
  namespace db {
    class result {
      public:
	result (const pqxx::result& r);
	result () = delete;

	const std::string& query () const;
	const std::string& table_name () const;

	size_t columns () const;
	const std::string& column_name (size_t column) const;
	std::vector<std::string> column_names () const;

	size_t rows () const;
	bool empty () const;

	bool check_cell (const std::string& column, size_t row) const;

	bool check_cell (size_t column, size_t row) const;

	template<typename T> size_t cell_elements (const T& column, size_t row) const;

	template<typename T> bool check_cell (const T& column, size_t row, size_t cell_element) const;

	template<typename T> T get (const std::string& column, size_t row) const;

	template<typename T> T get (size_t column, size_t row) const;

	template<typename T> T get (const std::string& column, size_t row, size_t cell_element) const;

	template<typename T> T get (size_t column, size_t row, size_t cell_element) const;

	void wipe ();

	template<typename T> void set (const std::string& column, size_t row, const T& value);

	template<typename T> void set (const std::string& column, size_t row, size_t cell_element, const T& value);

	result& set_table_name (const std::string &name);

      private:
	std::map<std::string, std::vector<std::pair<std::string, std::vector<std::string>>>> cells_;
	std::string query_;
	std::string table_name_;
    };
  }
}

inline const std::string& ds50::db::result::query () const { return query_; }

inline const std::string& ds50::db::result::table_name () const { return table_name_; }

inline size_t ds50::db::result::columns () const { return cells_.size (); }

inline size_t ds50::db::result::rows () const { return cells_.size () ? cells_.begin ()->second.size () : 0; }

inline bool ds50::db::result::empty () const { return !rows (); }


template<typename T> size_t ds50::db::result::cell_elements (const T& column, size_t row) const {
  if (!check_cell (column, row)) return 0;
  return cells_.find (column)->second[row].second.size ();
}

template<typename T> bool ds50::db::result::check_cell (const T& column, size_t row, size_t cell_element) const { 
  return (cell_elements (column, row) > cell_element); 
}

template<typename T> T ds50::db::result::get (const std::string& column, size_t row) const {
  if (!check_cell (column, row)) throw cet::exception("dbi::get") << "no cell (" << column << "," << row << ")" << (table_name_.size () ? " in table " : "") << table_name_;
  const auto &x = cells_.find (column)->second[row].first;
  try {
    return boost::lexical_cast<T>(x);
  } catch (boost::bad_lexical_cast& e) {
    throw cet::exception("dbi::get") << "bad cast in (" << column << "," << row << ")" << (table_name_.size () ? " in table " : "") << table_name_ << ": " << e.what ();
  }
}

template<typename T> T ds50::db::result::get (size_t column, size_t row) const {
  if (!check_cell (column, row)) throw cet::exception("dbi::get") << "no cell (" << column << "," << row << ")" << (table_name_.size () ? " in table " : "") << table_name_;
  auto c = cells_.begin ();
  std::advance (c, column);
  const auto &x = c->second[row].first;
  try {
    return boost::lexical_cast<T>(x);
  } catch (boost::bad_lexical_cast& e) {
    throw cet::exception("dbi::get") << "bad cast in (" << column << "," << row << ")" << (table_name_.size () ? " in table " : "") << table_name_ << ": " << e.what ();
  }
}

template<typename T> T ds50::db::result::get (const std::string& column, size_t row, size_t cell_element) const {
  if (!check_cell (column, row, cell_element)) throw cet::exception("dbi::get") << "no cell (" << column << "," << row << "," << cell_element << ")" << (table_name_.size () ? " in table " : "") << table_name_;
  const auto &x = cells_.find (column)->second[row].second[cell_element];
  try {
    return boost::lexical_cast<T>(x);
  } catch (boost::bad_lexical_cast& e) {
    throw cet::exception("dbi::get") << "bad cast in (" << column << "," << row << "," << cell_element << ")" << (table_name_.size () ? " in table " : "") << table_name_ << ": " << e.what ();
  }
}

template<typename T> T ds50::db::result::get (size_t column, size_t row, size_t cell_element) const {
  if (!check_cell (column, row, cell_element)) throw cet::exception("dbi::get") << "no cell (" << column << "," << row << "," << cell_element << ")" << (table_name_.size () ? " in table " : "") << table_name_;
  auto c = cells_.begin ();
  std::advance (c, column);
  const auto &x = c->second[row].second[cell_element];
  try {
    return boost::lexical_cast<T>(x);
  } catch (boost::bad_lexical_cast& e) {
    throw cet::exception("dbi::get") << "bad cast in (" << column << "," << row << "," << cell_element << ")" << (table_name_.size () ? " in table " : "") << table_name_ << ": " << e.what ();
  }
}

template<typename T> void ds50::db::result::set (const std::string& column, size_t row, const T& value) {
  auto x = cells_.find (column);
  if (x == cells_.end ()) throw cet::exception("dbi::set") << "column \"" << column << "\" does not exist" << (table_name_.size () ? " in table " : "") << table_name_;
  if (row >= x->second.size ()) x->second.resize (row + 1);
  try {
    x->second[row].first = boost::lexical_cast<std::string>(value);
  } catch (boost::bad_lexical_cast& e) {
    throw cet::exception("dbi::set") << "bad cast to string setting (" << column << "," << row << ")" << (table_name_.size () ? " in table " : "") << table_name_ << ": " << e.what ();
  }
  x->second[row].second.clear ();
}

template<typename T> void ds50::db::result::set (const std::string& column, size_t row, size_t cell_element, const T& value) {
  auto x = cells_.find (column);
  if (x == cells_.end ()) throw cet::exception("dbi::set") << "column " << column << " does not exist" << (table_name_.size () ? " in table " : "") << table_name_;
  if (row >= x->second.size ()) x->second.resize (row + 1);
  if (cell_element >= x->second[row].second.size ()) x->second[row].second.resize (cell_element + 1);
  try {
    x->second[row].second[cell_element] = boost::lexical_cast<std::string>(value);
  } catch (boost::bad_lexical_cast& e) {
    throw cet::exception("dbi::set") << "bad cast to string setting (" << column << "," << row << "," << cell_element << ")" << (table_name_.size () ? " in table " : "") << table_name_ << ": " << e.what ();
  }
  std::string v = "{";
  for (const auto &e: x->second[row].second) v += e + ",";
  v[v.size () - 1] = '}';
  x->second[row].first = v;
}

inline ds50::db::result& ds50::db::result::set_table_name (const std::string& name) { return (table_name_ = name), *this; }

#endif
