#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Principal/Run.h"
#include "art/Framework/Principal/SubRun.h"
#include "art/Persistency/Provenance/BranchType.h"
#include "darksidecore/Compression/Decoder.hh"
#include "darksidecore/Compression/Properties.hh"
#include "darksidecore/Compression/SymTable.hh"
#include "darksidecore/Data/CompressedV172x.hh"
#include "artdaq-core/Data/Fragment.hh"
#include "artdaq-core/Data/Fragments.hh"
#include "artdaq-core/Utilities/SimpleLookupPolicy.h"
#include "cpp0x/memory"
#include "fhiclcpp/ParameterSet.h"
#include "messagefacility/MessageLogger/MessageLogger.h"
#include "darksidecore/ArtServices/DBInterface.hh"

#include <cmath>
#include <iostream>
#include <string>
#include <sstream>

namespace ds50 {

  class DS50Decompression : public art::EDProducer
  {
  public:
    explicit DS50Decompression(fhicl::ParameterSet const & p);
    virtual ~DS50Decompression() { }

    virtual void produce(art::Event & e);
    virtual void beginRun(art::Run & r);
    virtual void endSubRun(art::SubRun & sr);
    virtual void endRun(art::Run & r);

  private:
    std::string compressed_label_;
    std::string inst_name_;
    bool use_db_for_table_;
    std::string db_symtable_name_;
    std::string table_file_;
    std::string table_file_path_;
    std::unique_ptr<Decoder> decode_;
    art::RunNumber_t current_run_;
  };

  static SymTable readAndSortTable(std::string const & fname,
                                   std::string const & fpath)
  {
    // 08-Apr-2013, KAB - added the ability to find the specified
    // table file in a list of paths specified explicitly or in an
    // environmental variable
    artdaq::SimpleLookupPolicy::ArgType argType =
      artdaq::SimpleLookupPolicy::ENV_VAR;
    if (getenv(fpath.c_str()) == nullptr) {
      if (fpath.find("/") == std::string::npos &&
          fpath.find(".") == std::string::npos) {
        setenv(fpath.c_str(), ".", 0);
      }
      else {
        argType = artdaq::SimpleLookupPolicy::PATH_STRING;
      }
    }
    artdaq::SimpleLookupPolicy lookup_policy(fpath, argType);
    std::string fullPath = fname;
    try {fullPath = lookup_policy(fname);}
    catch (...) {}

    SymTable t;
    readTable(fullPath.c_str(), t, Properties::count_max());
    return t;
  }

  DS50Decompression::DS50Decompression(fhicl::ParameterSet const & p)
    : compressed_label_(p.get<std::string>("compressed_label")),
      inst_name_(p.get<std::string>("instance_name")),
      use_db_for_table_(p.get<bool>("use_db_for_table", true)),
      db_symtable_name_(use_db_for_table_ ?
                        p.get<std::string>("symtable_name") :
                        "not_applicable"),
      table_file_(use_db_for_table_ ? "not_applicable" :
                  p.get<std::string>("table_file")),
      table_file_path_(p.get<std::string>("table_file_path", "DS50DAQ_CONFIG_PATH")),
      decode_(nullptr),
      current_run_(0)
  {
    produces<artdaq::Fragments>(inst_name_);

    if (! use_db_for_table_) {
      SymTable table(readAndSortTable(table_file_, table_file_path_));
      decode_.reset(new Decoder(table));
    }
  }

  void DS50Decompression::beginRun(art::Run &e) { 
    if (e.run () == current_run_) return;
    current_run_ = e.run ();

    if (use_db_for_table_) {
      SymTable t;
      try {
        art::ServiceHandle<ds50::DBInterface> dbi;
        const ds50::db::result & r = dbi->latest ("calib.compression_symtables", "");
        if (!r.check_cell (db_symtable_name_, 0)) throw cet::exception("DBError") << "no data for " << db_symtable_name_ << "@compression_symtables";
        std::istringstream f(r.get<std::string>(db_symtable_name_, 0));
        readTable(f, t, Properties::count_max());
      } catch (const std::runtime_error& e) {
        mf::LogError ("DBError") << "cannot read db";
        throw;
      }

      decode_ = std::unique_ptr<Decoder>(new Decoder(t));
    }
  }

#define ORIG_WAY 0
  void DS50Decompression::produce(art::Event & e)
  {
    art::Handle<CompressedV172x> handle;
    e.getByLabel(compressed_label_, inst_name_, handle);
    size_t len = handle->size();
    bool use_diffs = handle->getBias() > 0;

#if ORIG_WAY
    std::unique_ptr<artdaq::Fragments> prod(new artdaq::Fragments(len));
#else
    std::unique_ptr<artdaq::Fragments> prod(new artdaq::Fragments);
#endif

    for (size_t i = 0; i < len; ++i)
      {
#if ORIG_WAY
	artdaq::Fragment & newfrag = (*prod)[i];
	newfrag = (handle->headerOnlyFrag(i)); // Load in the header
	V172xFragment b_orig(newfrag);
	size_t cnt = ceil(b_orig.total_adc_values() * 
			  sizeof(adc_type))/sizeof(reg_type);
	newfrag.resize(cnt + V172xFragment::header_size_words()/2);
	V172xFragment b(newfrag);
	adc_type* new_begin = (adc_type*)(&*newfrag.dataBegin())+
			       (V172xFragment::header_size_words()*2);
	  

	std::cerr << "event=" << e.event()
		  << "dist=" 
		  << std::distance((adc_type const*)(new_begin),b.dataEnd())
		  << " bits=" << handle->fragmentBitCount(i)
		  << "\n";

	auto size_check __attribute__((unused))
	  (decode_.get ()(handle->fragmentBitCount(i),
		   handle->fragment(i).begin(),
		   new_begin,
		   b.dataEnd()));
	assert(size_check == b.total_adc_values());

#else
        // Load in the old header                                               
	artdaq::Fragment orig_frag = (handle->headerOnlyFrag(i));
        // size_t curr_payload_sz = orig_frag.dataSize();                       
        V172xFragment b_orig(orig_frag);
        size_t cnt = ceil(b_orig.total_adc_values() *
			  sizeof(adc_type))/sizeof(reg_type);
        orig_frag.resize(cnt + V172xFragment::header_size_words()/2);
        V172xFragment b(orig_frag);

	(*decode_)(handle->fragmentBitCount(i),
		handle->fragment(i).begin(),
		(adc_type*)(&*orig_frag.dataBegin())+
		(V172xFragment::header_size_words()*2),
		b.dataEnd(),
		use_diffs);
	
	prod->push_back(orig_frag);
#endif
      }
    e.put(std::move(prod), inst_name_);
  }

  void DS50Decompression::endSubRun(art::SubRun &) { }
  void DS50Decompression::endRun(art::Run &) { }

  DEFINE_ART_MODULE(DS50Decompression)
}

