////////////////////////////////////////////////////////////////////////
// Class:       DS50EventCheck
// Module Type: filter
// File:        DS50EventCheck_module.cc
// Description: Prints out information about each event.
////////////////////////////////////////////////////////////////////////

#include "art/Framework/Core/EDFilter.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"

#include "art/Utilities/Exception.h"
#include "darksidecore/Data/V172xFragment.hh"
#include "darksidecore/Data/V1495Fragment.hh"
#include "artdaq-core/Data/Fragments.hh"

#include <algorithm>
#include <cassert>
#include <cmath>
#include <fstream>
#include <iomanip>
#include <vector>
#include <iostream>

namespace ds50 {
  class DS50EventCheck;
}

class ds50::DS50EventCheck : public art::EDFilter {
public:
  explicit DS50EventCheck(fhicl::ParameterSet const & pset);
  virtual ~DS50EventCheck();

  virtual bool filter(art::Event & evt);

private:
  std::string raw_data_label_;
  std::string uncompressed_V1720_label_;
  size_t V1720_expected_adc_values_;
};


ds50::DS50EventCheck::DS50EventCheck(fhicl::ParameterSet const & pset) :
  raw_data_label_(pset.get<std::string>("raw_data_label")),
  uncompressed_V1720_label_(pset.get<std::string>("uncompressed_V1720_label")),
  V1720_expected_adc_values_(pset.get<size_t>("expected_adc_values_V1720"))
{
}

ds50::DS50EventCheck::~DS50EventCheck()
{
}

bool ds50::DS50EventCheck::filter(art::Event & evt)
{
  bool goodEvent = true;

  art::EventNumber_t eventNumber = evt.event();

  // ***********************
  // *** V1720 Fragments ***
  // ***********************

  // look for raw V1720 data first, but also check for uncompressed
  // V1720 data
  art::Handle<artdaq::Fragments> raw;
  evt.getByLabel(raw_data_label_, "V1720", raw);
  if (! raw.isValid()) {
    evt.getByLabel(uncompressed_V1720_label_, "V1720", raw);
  }

  if (raw.isValid()) {
    for (size_t idx = 0; idx < raw->size(); ++idx) {
      const auto& frag((*raw)[idx]);

      if (frag.sequenceID() != eventNumber) {
        std::cout << "*** ERROR: the fragment sequence number ("
                  << frag.sequenceID() << ") in fragment "
                  << frag.fragmentID()
                  << " does not match event number ("
                  << eventNumber << ")." << std::endl;
        goodEvent = false;
      }

      V172xFragment bb(frag);
      if ((bb.event_counter() & 0xffffff) != (eventNumber & 0xffffff)) {
        std::cout << "*** ERROR in event " << eventNumber
                  << ": the least significant 24 bits of the V1720 "
                  << "event counter in fragment " << frag.fragmentID()
                  << " (0x" << std::hex
                  << bb.event_counter() << ") do not match the "
                  << "least significant 24 bits of the event number "
                  << "(0x" << eventNumber << ")." << std::dec
                  << std::endl;
        goodEvent = false;
      }

      if (bb.total_adc_values() != V1720_expected_adc_values_) {
        std::cout << "*** ERROR in event " << eventNumber
                  << ": the total number of ADC values ("
                  << bb.total_adc_values() << ") in fragment "
                  << frag.fragmentID()
                  << " does not match the "
                  << "expected number ("
                  << V1720_expected_adc_values_ << ")."
                  << std::endl;
        goodEvent = false;
      }
    }
  }
  else {
    std::cout << "*** ERROR in event " << eventNumber
              << ": failed to find raw or uncompressed V1720 data."
              << std::endl;
    goodEvent = false;
  }

  // ***********************
  // *** V1495 Fragments ***
  // ***********************

  evt.getByLabel(raw_data_label_, "V1495", raw);

  if (raw.isValid()) {
    for (size_t idx = 0; idx < raw->size(); ++idx) {
      const auto& frag((*raw)[idx]);

      V1495Fragment bb(frag);
      if ((bb.trigger_counter() & 0xfff) != (eventNumber & 0xfff)) {
        std::cout << std::showbase << "*** ERROR in event " << eventNumber
                  << ": the least significant 12 bits of the V1495 "
                  << "trigger counter (" << std::hex
                  << bb.trigger_counter () << ") do not match the "
                  << "least significant 12 bits of the event number "
                  << "(0x" << eventNumber << ")." << std::dec
                  << std::endl;
        goodEvent = false;
      }
    }
  }
  else {
    std::cout << "*** ERROR in event " << eventNumber
              << ": failed to find V1495 data."
              << std::endl;
    goodEvent = false;
  }

  return goodEvent;
}

DEFINE_ART_MODULE(ds50::DS50EventCheck)
