#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Principal/Run.h"
#include "art/Framework/Principal/SubRun.h"
#include "art/Persistency/Provenance/BranchType.h"
#include "art/Persistency/Provenance/EventID.h"
#include "darksidecore/Compression/Encoder.hh"
#include "darksidecore/Compression/Codes.hh"
#include "darksidecore/Compression/Properties.hh"
#include "darksidecore/Compression/StatsKeeper.hh"
#include "darksidecore/Compression/SymTable.hh"
#include "darksidecore/Data/V172xFragment.hh"
#include "darksidecore/Data/CompressedV172x.hh"
#include "artdaq-core/Data/Fragment.hh"
#include "artdaq-core/Data/Fragments.hh"
#include "cpp0x/memory"
#include "fhiclcpp/ParameterSet.h"
#include "messagefacility/MessageLogger/MessageLogger.h"

#include <cstdlib>
#include <iostream>
#include <list>
#include <ostream>
#include <fstream>
#include <sstream>
#include <string>
#include <utility>

using std::vector;
using namespace std;


namespace ds50 {

  class DS50ExpCompress : public art::EDProducer 
  {
  public:
    explicit DS50ExpCompress(fhicl::ParameterSet const & p);
    virtual ~DS50ExpCompress() { }
    
    virtual void produce(art::Event & e);
    virtual void endSubRun(art::SubRun & sr);
    virtual void endRun(art::Run & r);

  private:
    std::string const mod_label_;
    std::string const raw_label_;
    std::string const inst_name_;
    std::string const algo_;
    CompressedV172x::Algo_t algo_type_;
    unsigned bias_;
    int bits_;
    bool want_bins_;
    bool perf_print_;
    SymTable table_;
    StatsKeeper stats_;
  };
  
  DS50ExpCompress::DS50ExpCompress(fhicl::ParameterSet const & p):
  mod_label_(p.get<std::string>("module_label")),
    raw_label_(p.get<std::string>("raw_label","daq")),
    inst_name_(p.get<std::string>("instance_name")),
    algo_(p.get<std::string>("algorithm","pod")),
    algo_type_(algo_=="pod"?CompressedV172x::AlgoPod 
	       : algo_=="subexp0"?CompressedV172x::AlgoSubexp0 
	       : algo_=="subexp1" ? CompressedV172x::AlgoSubexp1 
	       : CompressedV172x::AlgoUnknown),
    bias_(p.get<unsigned>("bias",0)),
    bits_(inst_name_=="V1724" ? 14 : inst_name_=="V1720" ? 12 : 0),
    want_bins_(p.get<bool>("want_bins",false)),
    perf_print_(p.get<bool>("perf_print",false)),
    table_(),
    stats_()
  {
    produces<CompressedV172x>(inst_name_);
    typedef Code (*Func)(code_type);
    Func comp_func = 0;
    
    switch(algo_type_)
      {
      case CompressedV172x::AlgoPod: comp_func=&pod; break;
      case CompressedV172x::AlgoSubexp0: comp_func = &subexp<0>; break;
      case CompressedV172x::AlgoSubexp1: comp_func = &subexp<1>; break;
      default: throw std::runtime_error("valid algos are pod, subexp0, subexp1");
      }

    generateTable(comp_func, table_, 50);
    ds50::writeTable((algo_+"_temp_out_exp_table.txt").c_str(),table_);
  }

  void DS50ExpCompress::produce(art::Event & e)
  {
    art::Handle<artdaq::Fragments> handle;
    e.getByLabel(raw_label_, inst_name_, handle);
    std::unique_ptr<CompressedV172x> prod(new CompressedV172x(*handle));
    prod->setMetadata(bits_, algo_type_, bias_);
    size_t const len = handle->size();

    for (size_t i = 0; i < len; ++i) 
      {
	artdaq::Fragment const & frag = (*handle)[i];
	V172xFragment b(frag);
	// start of payload is the DS50 header
	auto adc_start = b.dataBegin();
	auto adc_end   = b.dataEnd();

	// make sure the destination is big enough
	size_t big_enough = (size_t)(frag.dataSize() * 2.0);
	prod->fragment(i).resize(big_enough);
	unsigned long bit_count  = rleAndCompress(bits_,adc_start, adc_end, prod->fragment(i), table_, bias_);
	// shrink destination to the actual size
	prod->fragment(i).resize(std::ceil(bit_count / (8.0 * sizeof(DataVec::value_type))));
	prod->setFragmentBitCount(i, bit_count);
	
	stats_.put((*handle)[i].dataSize(), prod->fragment(i).size());

	if(perf_print_)
	  {
	    cout << "ExpCompress" << " "
		 << mod_label_ << " "
		 << inst_name_ << " "
		 << raw_label_ << " "
		 << algo_ << " " 
		 << e.event() << " "
		 << i << " "
		 << ((double)(*handle)[i].dataSize() / (double)prod->fragment(i).size()) << " "
		 << ((double)prod->fragment(i).size() / (double)(*handle)[i].dataSize()) << " "
		 << "\n";
	  }

      }
    // need to include bias and algorithm name to the product
    e.put(std::move(prod),inst_name_);

  }

  void DS50ExpCompress::endSubRun(art::SubRun &) { }

  void DS50ExpCompress::endRun(art::Run & r)
  {
    mf::LogInfo("STATS") << stats_;

    if(want_bins_)
      {
	pid_t pid = getpid();
	ostringstream ostr;
	ostr << "bins" << "_" << pid << "_" << r.run() << ".txt";
	ofstream ost(ostr.str().c_str());
	for(auto const& x : stats_.bins()) { ost << x << "\n"; }
	ost.close();
      }
  }

  DEFINE_ART_MODULE(DS50ExpCompress)
}

