#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Principal/Run.h"
#include "art/Framework/Principal/SubRun.h"
#include "art/Persistency/Provenance/BranchType.h"
#include "darksidecore/Compression/Properties.hh"
#include "darksidecore/Compression/ExpDecode.hh"
#include "darksidecore/Data/CompressedV172x.hh"
#include "artdaq-core/Data/Fragment.hh"
#include "artdaq-core/Data/Fragments.hh"
#include "cpp0x/memory"
#include "fhiclcpp/ParameterSet.h"
#include "messagefacility/MessageLogger/MessageLogger.h"

#include <cmath>
#include <iostream>
#include <string>

namespace ds50 {

  class DS50ExpDecomp : public art::EDProducer 
  {
  public:
    explicit DS50ExpDecomp(fhicl::ParameterSet const & p);
    virtual ~DS50ExpDecomp() { }

    virtual void produce(art::Event & e);
    virtual void endSubRun(art::SubRun & sr);
    virtual void endRun(art::Run & r);

  private:
    std::string comp_label_;
    std::string inst_name_;
  };

  DS50ExpDecomp::DS50ExpDecomp(fhicl::ParameterSet const & p)
    : comp_label_(p.get<std::string>("compressed_label","daq")),
      inst_name_(p.get<std::string>("instance_name"))
  {
    produces<artdaq::Fragments>(inst_name_);
  }

  void DS50ExpDecomp::produce(art::Event & e)
  {
    art::Handle<CompressedV172x> handle;
    e.getByLabel(comp_label_, inst_name_, handle);
    size_t len = handle->size();

    std::unique_ptr<artdaq::Fragments> prod(new artdaq::Fragments);
    ADCCountVec tmp_out;

    for (size_t i = 0; i < len; ++i) 
      {
	// Load in the old header
	artdaq::Fragment orig_frag = (handle->headerOnlyFrag(i)); 
	// size_t curr_payload_sz = orig_frag.dataSize();
	V172xFragment b_orig(orig_frag);
	size_t cnt = ceil(b_orig.total_adc_values() * sizeof(adc_type))/sizeof(reg_type);
	orig_frag.resize(cnt + V172xFragment::header_size_words()/2);
	V172xFragment b(orig_frag);

	// decompress into this range:
	// b_new.dataBegin(), b_new.dataEnd()
	// from range:
	// handle->fragment(i).begin(), handle->fragmentBitCount(i)
	// easier: orig_frag.dataBegin()+V172xFragment::header_size_words() as destination?

	tmp_out.clear();
	
	decodePod(handle->getAdcBits(), handle->fragmentBitCount(i),
		  &*handle->fragment(i).begin(),
		  (adc_type*)(&*orig_frag.dataBegin())+
		  (V172xFragment::header_size_words()*2));
	
	// the above '2' should be adcs_per_word, but that is protected info

	prod->push_back(orig_frag);
      }

    e.put(std::move(prod), inst_name_);
  }

  void DS50ExpDecomp::endSubRun(art::SubRun &) { }
  void DS50ExpDecomp::endRun(art::Run &) { }

  DEFINE_ART_MODULE(DS50ExpDecomp)
}

