#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Principal/Run.h"
#include "art/Framework/Principal/SubRun.h"
#include "art/Persistency/Provenance/BranchType.h"
#include "art/Persistency/Provenance/EventID.h"
#include "darksidecore/Compression/Codes.hh"
#include "darksidecore/Compression/Properties.hh"
#include "darksidecore/Compression/HuffmanTable.hh"
#include "darksidecore/Compression/SymTable.hh"
#include "darksidecore/Compression/SymProb.hh"
#include "darksidecore/Data/V172xFragment.hh"
#include "artdaq-core/Data/Fragment.hh"
#include "artdaq-core/Data/Fragments.hh"
#include "cpp0x/memory"
#include "fhiclcpp/ParameterSet.h"
#include "messagefacility/MessageLogger/MessageLogger.h"

#include <cstdlib>
#include <iostream>
#include <list>
#include <ostream>
#include <fstream>
#include <sstream>
#include <string>
#include <utility>

using std::vector;
using namespace std;

namespace {

}

namespace ds50 {

  class DS50ProbCalculator : public art::EDAnalyzer
  {
  public:
    explicit DS50ProbCalculator(fhicl::ParameterSet const & p);
    virtual ~DS50ProbCalculator() { }
    
    virtual void analyze(art::Event const & e);
    virtual void endSubRun(art::SubRun const & sr);
    virtual void endRun(art::Run const & r);

  private:
    std::string const mod_label_;
    std::string const raw_label_;
    std::string const inst_name_;
    size_t bits_;
    ProbCalculator calc_reg_;
    ProbCalculator calc_diff_;
    std::string probs_file_reg_;
    std::string probs_file_diff_;
    std::string huff_file_reg_;
    std::string huff_file_reg_rev_;
    std::string huff_file_diff_;
    std::string huff_file_diff_rev_;
  };
  
  DS50ProbCalculator::DS50ProbCalculator(fhicl::ParameterSet const & p):
    art::EDAnalyzer(p),
    mod_label_(p.get<std::string>("module_label")),
    raw_label_(p.get<std::string>("raw_label","daq")),
    inst_name_(p.get<std::string>("instance")),
    bits_(inst_name_=="V1724" ? 14 : inst_name_=="V1720" ? 12 : 0),
    calc_reg_(bits_,false),
    calc_diff_(bits_,true),
    probs_file_reg_(mod_label_+raw_label_+inst_name_+"_probs_reg.txt"),
    probs_file_diff_(mod_label_+raw_label_+inst_name_+"_probs_diff.txt"),
    huff_file_reg_(mod_label_+raw_label_+inst_name_+"_huff_reg.txt"),
    huff_file_reg_rev_(mod_label_+raw_label_+inst_name_+"_huff_reg_rev.txt"),
    huff_file_diff_(mod_label_+raw_label_+inst_name_+"_huff_diff.txt"),
    huff_file_diff_rev_(mod_label_+raw_label_+inst_name_+"_huff_diff_rev.txt")
  {
    if(bits_==0)
      throw std::runtime_error("invalid instance name, must be V1720 or V1724");
  }

  void DS50ProbCalculator::analyze(art::Event const & e)
  {
    art::Handle<artdaq::Fragments> handle;
    e.getByLabel(raw_label_, inst_name_, handle);
    size_t const len = handle->size();

    for (size_t i = 0; i < len; ++i) 
      {
	artdaq::Fragment const & frag = (*handle)[i];
	V172xFragment b(frag);
	// start of payload is the DS50 header
	auto adc_start = b.dataBegin();
	auto adc_end   = b.dataEnd();

	calc_reg_.apply( adc_start, adc_end );
	calc_diff_.apply( adc_start, adc_end );

#if 0	
	cout << "ProbCalculator:sum"
	     << " " << mod_label_
	     << " " << raw_label_
	     << " " << inst_name_
	     << " " << e.event()
	     << " " << i
	     << "\n";
#endif
      }
  }

  void DS50ProbCalculator::endSubRun(art::SubRun const &) { }

  void DS50ProbCalculator::endRun(art::Run const & )
  {
    SymsVec syms_reg;
    calc_reg_.get(syms_reg);
    SymsVec syms_diff;
    calc_diff_.get(syms_diff);

#if 1
    HuffmanTable ht_reg(syms_reg);
    ht_reg.writeTable( huff_file_reg_.c_str() );
    ht_reg.writeTableReversed(  huff_file_reg_rev_.c_str() );
#endif
#if 1
    HuffmanTable ht_diff(syms_diff);
    ht_diff.writeTable( huff_file_diff_.c_str() );
    ht_diff.writeTableReversed(  huff_file_diff_rev_.c_str() );
#endif

#if 1
    std::ofstream prob_str_d(probs_file_diff_.c_str());
    if(!prob_str_d) throw std::runtime_error("no open of probs file");
    for_each(syms_diff.begin(), syms_diff.end(),
	     [&](SymsVec::value_type & v)
	     { prob_str_d << v << "\n"; });

    std::ofstream prob_str_r(probs_file_reg_.c_str());
    if(!prob_str_r) throw std::runtime_error("no open of probs file");
    for_each(syms_reg.begin(), syms_reg.end(),
	     [&](SymsVec::value_type & v)
	     { prob_str_r << v << "\n"; });
#endif
  }

  DEFINE_ART_MODULE(DS50ProbCalculator)
}

